//
// JTypist v3.1 - Java typing tutor program for UNIX systems
// Copyright (C) 2000  Simon Baldwin (simonb@sco.com)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//
// Xpm class - class for handling XPM graphics
//
import java.awt.Color;
import java.awt.image.ImageProducer;
import java.awt.image.MemoryImageSource;
import java.util.StringTokenizer;
import java.util.Hashtable;
import java.util.ResourceBundle;

/**
 * Static methods for parsing an XPM graphic into a ImageProducer.
 *
 * Xpm images may be sent directly, as character arrays, or, a properties
 * file name may be given.  If a properties file is used, the file should have
 * the format:
 * <p><pre>
 * XPM = \n\
 * "width height colours chars_per_colour"\n\
 * "ctag 'c' #colour_spec"\n\
 * ...
 * "data line 1"\n\
 * ...
 * "data line N"
 * </pre><p>
 * Here are a couple of Unix shell scripts to convert real something.xpm
 * files into XpmSomething.properties files, and vice-versa:
 * <p><pre>
 *     #!/bin/sh
 *     # x2p - xpm to properties filter
 *     sed -e '/^\/\*.*$/d' \
 *       -e 's;^static char .*$;XPM = \\n\\;' \
 *       -e 's;,$;\\n\\;' \
 *       -e '/^};$/d' -e 's;}\;$;;'
 * </pre><p><pre>
 *     #!/bin/sh
 *     # p2x - properties to xpm filter
 *     echo -e "\057\052 XPM \052\057"
 *     sed -e '/^#.*$/d' \
 *       -e 's;^XPM[[:space:]]*=[[:space:]]*\\n\\;static char * xpm[] = {;' \
 *       -e 's;\\n\\$;,;'
 *     echo "};"
 * </pre><p>
 * Note that this class makes no attempt to recover from badly-formed Xpm
 * images.
 */
public class Xpm {

	/**
	 * Return a MemoryImageProducer initialized with data from an XPM
	 * held in a string array.
	 */
	public static ImageProducer get_image_producer( String[] xpm ) {

		// line pointer through the XPM definition
		int xpm_line	= 0;

		// find the Xpm dimensions
		StringTokenizer t = new StringTokenizer( xpm[ xpm_line++ ] );
		int width		= Integer.parseInt( t.nextToken() );
		int height		= Integer.parseInt( t.nextToken() );

		// find the Xpm number of colours, and make a table for them
		int num_colours		= Integer.parseInt( t.nextToken() );
		Hashtable colours_by_ctag = new Hashtable( num_colours );
		int chars_per_colour	= Integer.parseInt( t.nextToken() );

		// analyse the colour map presented for this image
		for ( int ccount = 0; ccount < num_colours;
							ccount++,xpm_line++ ) {

			// get the colour tag string
			String ctag = xpm[ xpm_line ].substring( 0,
							chars_per_colour );

			// get the colour definition
			StringTokenizer c = new StringTokenizer(
					xpm[ xpm_line ].substring(
							chars_per_colour ));
			while ( ! c.nextToken().equals( "c" ))
				;
			String cstring = c.nextToken();

			// turn the #FFFFFFFFFFFF, #FFFFFF, or None into
			// something decodable, then decode it into a colour;
			// if the colour is the special "None" value, then
			// we don't assign any colour to the tags table
			if ( ! cstring.equalsIgnoreCase( "None" )) {

				// find the colour the string represents
				Color colour;
				if ( cstring.charAt( 0 ) != '#' ) {
					// colour by name, e.g. "green"
					colour = Colour.find_colour( cstring );
				}
				else if ( cstring.length() > 7 ) {
					// colour by 48-bit definition; take
					// just the high-order byte of each
					colour = Colour.find_colour( "0x" +
						cstring.substring( 1,  3 ) +
						cstring.substring( 5,  7 ) +
						cstring.substring( 9, 11 ));
				}
				else {
					// colour by 24-bit definition; use
					// this one as is
					colour = Colour.find_colour( "0x" +
						cstring.substring( 1 ));
				}

				// add the colour to the lookup table
				colours_by_ctag.put( ctag, colour );
			}
		}

		// now go through the image, and build a colour pixel array
		int[] pixels = new int[ width * height ];
		for ( int y = 0; y < height; y++,xpm_line++ ) {
			for ( int x = 0; x < width; x++ ) {

				// extract the colour tag substring
				String ctag = xpm[ xpm_line ].substring(
						chars_per_colour * x,
						chars_per_colour * x
							+ chars_per_colour );

				// lookup the colour, and assign it; if the
				// lookup returns null, this means the colour
				// was "None", and we make this transparent
				Color c = (Color)colours_by_ctag.get( ctag );
				pixels[ y * height + x ] = ( c != null )
						? ( 0xff         << 24 )
						+ ( c.getRed()   << 16 )
						+ ( c.getGreen() << 8 )
						+   c.getBlue()
						: 0x00 << 24;
			}
		}

		// return the MemoryImageProducer for this pixels array
		return ( new MemoryImageSource( width, height,
							pixels, 0, width ) );
	}

	/**
	 * Return a MemoryImageProducer initialized with data from an XPM
	 * resource file.
	 */
	public static ImageProducer get_image_producer( String name ) {

		// locate the property resource bundle, and within it
		// locate the XPM resource
		ResourceBundle bundle = ResourceBundle.getBundle(
								"Xpm" + name );
		String xpm_resource = bundle.getString( "XPM" );

		// break the XPM string into pieces based on newline, and
		// put each piece into an xpm array element
		StringTokenizer tokenizer = new StringTokenizer(
							xpm_resource, "\n" );
		String xpm[] = new String[ tokenizer.countTokens() ];
		for ( int xpm_line = 0; tokenizer.hasMoreTokens(); xpm_line++ ){

			// get the next xpm line
			String line = tokenizer.nextToken();

			// strip off leading and trailing quotes; these are
			// left on in the conversion from .xpm format
			while ( line.charAt( 0 ) == '"' )
				line = line.substring( 1 );
			while ( line.charAt( line.length() - 1 ) == '"' )
				line = line.substring( 0, line.length() - 1 );

			// add the line to the array
			xpm[ xpm_line ] = line;
		}

		// return the ImageProducer from this array
		return ( get_image_producer( xpm ));
	}

	/**
	 * Class test code.
	 */
	public static void main( String args[] ) {
		System.out.println( "*** CLASS TEST CODE ***" );

		System.out.println( get_image_producer( "PenguinA48x48" ));
	}
}
